
import sys
import os
import hashlib
import struct
import array

import wx
import wx.lib.mixins.listctrl



class ShortTextSource:
    
    def __init__(self, filename, parser):
        txt = file(filename, 'r').read()
        s = txt.split("\n\n\n")

        self.textlist = map(parser, s)
        self.textdict = dict(self.textlist)

    def __len__(self): return len(self.textlist)
    def __contains__(self, key): return key in self.textdict
    def __getitem__(self, key): return self.textdict[key]
    def __setitem__(self, key): raise TypeError
    def __delitem__(self, key): raise TypeError
    def __iter__(self): return iter(self.textlist)
    
    @staticmethod
    def parseBlock(s):
        id, text = s.split("\n",1)
        return (id, text)

    @staticmethod
    def parseBlock2(s):
        id, text = s.split("\n",1)
        id, brief = id.split(" ",1)
        return (int(id), (brief, text))



class SelectListBox(wx.ListCtrl, wx.lib.mixins.listctrl.CheckListCtrlMixin):
    
    def __init__(self, parent, id, pos=wx.DefaultPosition, size=wx.DefaultSize):
        wx.ListCtrl.__init__(self, parent, id, pos, size, style=wx.LC_REPORT|wx.LC_NO_HEADER|wx.LC_SINGLE_SEL)
        wx.lib.mixins.listctrl.CheckListCtrlMixin.__init__(self)
        self.InsertColumn(0, "")
        self.Bind(wx.EVT_SIZE, self.onSize)
        self.itemmap = {}
        
    def onSize(self, e):
        self.SetColumnWidth(0, self.GetClientSize().width-8)
        e.Skip(True)
        
    def AddItems(self, items):
        for i in items:
            newitem = wx.ListItem()
            idx, txt, sel = i
            img = 0

            if idx < 0:
                # Remove icon and attempt to centre text
                c = max(0, (40-len(txt))/2)
                txt = c*' ' + txt
                img = -1

            if sel:
                # Set checked
                img = 1

            listidx = self.InsertImageStringItem(self.GetItemCount(), txt, img)
            self.itemmap[listidx] = idx

    def GetAllCheckedData(self):
        sel = {}
        for i in range(self.GetItemCount()):
            if self.itemmap[i] > 0:
                sel[self.itemmap[i]] = self.IsChecked(i)
        return sel

    def GetSelectedData(self):
        idx = self.GetFirstSelected()
        if idx != -1:
            return self.itemmap[idx]
        else:
            return None

    def UncheckAll(self):
        for i in range(self.GetItemCount()):
            self.CheckItem(i, False)
            
        
        
class MainWindow(wx.Frame):

    def __init__(self, app):
        self.app = app
        self.loadStrings()
        wx.Frame.__init__(self, None, wx.ID_ANY, self.strmsg['sTitle'])
        self.setupWindow()

        if self.app.isInstalled():
            self.txtPatchInfo.SetValue(self.strmsg['sWelcome'])
            selected = self.app.getInstalled()
        else:
            self.txtPatchInfo.SetValue(self.strmsg['sFirstWelcome'])
            selected = self.app.getDefaultInstall()
            
        lst = map(lambda s: (s[0], s[1][0], selected.get(s[0], False)), self.strdesc.textlist)
        self.lstPatchSelector.AddItems(lst)

    def setupWindow(self):
        self.SetBackgroundColour(wx.SystemSettings.GetColour(wx.SYS_COLOUR_3DFACE))
        
        self.txtSelectorTitle = wx.StaticText(self, -1, self.strmsg['sSelectorTitle'])
        self.txtMessageTitle = wx.StaticText(self, -1, self.strmsg['sMessageTitle'])
        self.pnlLogo = wx.Panel(self, -1)
        self.imgLogo = wx.Image('mcpatch/logo_large.png', wx.BITMAP_TYPE_PNG)
        self.bmpLogo = wx.StaticBitmap(self.pnlLogo, bitmap=wx.BitmapFromImage(self.imgLogo))
        
        self.lstPatchSelector = SelectListBox(self, -1, size=(210,400))
        self.txtPatchInfo = wx.TextCtrl(self, -1, "", style=wx.TE_READONLY|wx.TE_MULTILINE|wx.TE_NO_VSCROLL)
        self.btnApply = wx.Button(self, -1, self.strmsg['sBtnApply'], size=(90,60))
        self.btnUninstall = wx.Button(self, -1, self.strmsg['sBtnUninstall'], size=(90,60))

        self.Bind(wx.EVT_CLOSE, self.onClose)
        self.btnApply.Bind(wx.EVT_BUTTON, self.onPatch)
        self.btnUninstall.Bind(wx.EVT_BUTTON, self.onUninstall)
        self.lstPatchSelector.Bind(wx.EVT_LIST_ITEM_SELECTED, self.onSelectPatch)
        
        self.sizer0 = wx.BoxSizer(wx.VERTICAL)
        self.sizer1 = wx.BoxSizer(wx.HORIZONTAL)
        self.sizer2 = wx.BoxSizer(wx.HORIZONTAL)
        self.sizer3 = wx.BoxSizer(wx.VERTICAL)
        self.sizer4 = wx.BoxSizer(wx.VERTICAL)

        self.sizer0.Add(self.pnlLogo, 0, wx.ALIGN_CENTRE_HORIZONTAL)
        self.sizer0.AddSpacer(8, 8)
        self.sizer0.Add(self.sizer1, 6, wx.EXPAND)
        self.sizer0.AddSpacer(8, 8)
        self.sizer0.Add(self.sizer2, 0, wx.EXPAND)
        self.sizer1.AddSpacer(8, 8)
        self.sizer1.Add(self.sizer3, 2, wx.EXPAND)
        self.sizer1.AddSpacer(8, 8)
        self.sizer1.Add(self.sizer4, 3, wx.EXPAND)
        self.sizer1.AddSpacer(8, 8)
        self.sizer2.Add(self.btnApply, 5)
        self.sizer2.Add(self.btnUninstall, 0)
        self.sizer3.Add(self.txtSelectorTitle, 0, wx.EXPAND)
        self.sizer3.Add(self.lstPatchSelector, 1, wx.EXPAND)
        self.sizer4.Add(self.txtMessageTitle, 0, wx.EXPAND)
        self.sizer4.Add(self.txtPatchInfo, 1, wx.EXPAND)

        self.SetSizer(self.sizer0)
        self.sizer0.Fit(self)

    def onPatch(self, e):
        self.btnApply.Enable(False)
        self.btnUninstall.Enable(False)
        
        ids = self.lstPatchSelector.GetAllCheckedData()
        self.app.onPatch(ids)

        self.btnApply.Enable(True)
        self.btnUninstall.Enable(True)
        
    def onUninstall(self,e):
        if self.app.onUninstall():
            self.lstPatchSelector.UncheckAll()

    def onSelectPatch(self, e):
        i = self.lstPatchSelector.GetSelectedData()
        if i:
            txt = self.strdesc[i][1]
            if txt != 'None':
                self.txtPatchInfo.SetValue("\n" + txt)

    def onClose(self, e):
        self.app.onClose()
        e.Skip(True)

    def loadStrings(self):
        self.strdesc = ShortTextSource('mcpatch/describe.txt', ShortTextSource.parseBlock2)
        self.strmsg = ShortTextSource('mcpatch/message.txt', ShortTextSource.parseBlock)

    def displayLog(self, log):
        txt = "\n".join(map(self.logSubstitute, log))
        self.txtPatchInfo.SetValue(txt)
        self.txtPatchInfo.Update()

    def logSubstitute(self, s):
        if s in self.strmsg:
            return self.strmsg[s]
        else:
            return s


        
class MCPatchApp(wx.App):
    
    def __init__(self):
        wx.App.__init__(self)
        self.loadInstalled()
        self.mergeInstallDefaults()
        self.mwname = 'Morrowind.exe'
        self.mwbackup = 'Morrowind.Original.exe'
        self.dllname = 'Text.dll'
        self.dllbackup = 'Text.Original.dll'
        self.fontname = 'Data Files/Fonts/Century_gothic_font_regular.fnt'

        self.frame = MainWindow(self)
        self.frame.Show(True)
        self.MainLoop()

    def loadInstalled(self):
        self.installed = {}
        if os.path.exists('mcpatch/installed'):
            f = file('mcpatch/installed', 'rb')
            s = f.read()
            f.close()

            if s[:4] == 'MCP2':
                y = array.array('l', s[4:])
                self.installed = dict(map(lambda x: (abs(x), x >= 0), y))
            else:
                y = array.array('l', s)
                self.installed = dict(map(lambda x: (x, True), y))

    def saveInstalled(self):
        y = map(lambda x: x if self.installed[x] else -x, self.installed.iterkeys())
        y = array.array('l', y)

        f = file('mcpatch/installed', 'wb')
        f.write('MCP2')
        y.tofile(f)
        f.close()
    
    def getInstalled(self):
        return self.installed

    def getDefaultInstall(self):
        f = file('mcpatch/default', 'rb')
        s = f.read()
        f.close()

        if s[:4] == 'MCP2':
            y = array.array('l', s[4:])
            return dict(map(lambda x: (abs(x), x >= 0), y))
        else:
            return {}

    def mergeInstallDefaults(self):
        x = self.getDefaultInstall()
        x.update(self.installed)
        self.installed = x
        
    def isInstalled(self):
        return 0 in self.installed
        
    def onClose(self):
        self.saveInstalled()
        
    def onPatch(self, selected):
        o = array.array('B')

        # Clear output pane
        self.clearLog()

        # Always apply infrastructure (id 0)
        selected[0] = True

        # Detect Morrowind and make backup if necessary
        self.addLog('sPatchStart')

        if not os.path.exists(self.mwname):
            self.addLog('sMorrowindNotFound')
            self.addLog('sPatchFailed')
            return False

        try:
            if os.path.exists(self.mwbackup):
                f = file(self.mwbackup, 'rb')
                o.fromstring(f.read())
                f.close()
            else:
                self.addLog('sPatchBackup')
                os.rename(self.mwname, self.mwbackup)
                f = file(self.mwbackup, 'rb')
                o.fromstring(f.read())
                f.close()
                f2 = file(self.mwname, 'wb')
                o.tofile(f2)
                f2.close()
        except IOError, inst:
            self.addLog('sIOError')
            self.addLog(inst.strerror)
            self.addLog('sPatchFailed')
            return False
            
        # Check hash and find available patch
        d = 'mcpatch/' + hashlib.sha256(o).hexdigest()

        if not os.path.exists(d):
            try:
                os.unlink(self.mwname)
                os.rename(self.mwbackup, self.mwname)
            except IOError, inst:
                self.addLog('sIOError')
                self.addLog(inst.strerror)
            self.addLog('sPatchNotAvailable')
            self.addLog('sPatchFailed')
            return False

        # Patch Russian dll if it's there
        if os.path.exists(self.dllname):
            if not self.onPatchRus(selected, d):
                self.addLog('sPatchFailed')
                return False

        # Patch font for Polish version
        if os.path.exists(self.fontname):
            if not self.onPatchFont(selected, d):
                self.addLog('sPatchFailed')
                return False

        # Patch main file
        if not self.patchFile(self.mwname, d + '/patch', o, selected):
            self.addLog('sPatchFailed')
            return False

        # Patch BSA dates
        if not self.patchBSAs('Data Files'):
            self.addLog('sPatchFailed')
            return False
        
        # Report done
        self.addLog('sPatchSuccess')
        self.installed = selected
        return True


    def onPatchRus(self, selected, patchdir):
        o = array.array('B')

        # Remove backup if new dll version is installed, just compare sizes
        try:
            if os.path.exists(self.dllname) and os.path.exists(self.dllbackup):
                if os.stat(self.dllname).st_size != os.stat(self.dllbackup).st_size:
                    self.addLog('sPatchRusChangedDLL')
                    os.unlink(self.dllbackup)
        except IOError, inst:
            self.addLog('sIOError')
            self.addLog(inst.strerror)
            return False
        
        # Backup dll if necessary
        try:
            if os.path.exists(self.dllbackup):
                f = file(self.dllbackup, 'rb')
                o.fromstring(f.read())
                f.close()
            else:
                self.addLog('sPatchRusBackup')
                os.rename(self.dllname, self.dllbackup)
                f = file(self.dllbackup, 'rb')
                o.fromstring(f.read())
                f.close()
                f2 = file(self.dllname, 'wb')
                o.tofile(f2)
                f2.close()
        except IOError, inst:
            self.addLog('sIOError')
            self.addLog(inst.strerror)
            return False
            
        # Check hash and find available patch
        d = patchdir + '/' + hashlib.sha256(o).hexdigest()

        if not os.path.exists(d):
            try:
                os.unlink(self.dllname)
                os.rename(self.dllbackup, self.dllname)
            except IOError, inst:
                self.addLog('sIOError')
                self.addLog(inst.strerror)
            self.addLog('sPatchRusNotAvailable')
            return False

        # Patch dll
        return self.patchFile(self.dllname, d + '/patch', o, selected)
        
    def onPatchFont(self, selected, patchdir):
        o = array.array('B')

        # Read font to confirm hash
        try:
            if os.path.exists(self.fontname):
                f = file(self.fontname, 'rb')
                o.fromstring(f.read())
                f.close()
            else:
                return False
        except IOError, inst:
            self.addLog('sIOError')
            self.addLog(inst.strerror)
            return False
            
        # Check hash and find available patch
        d = patchdir + '/' + hashlib.sha256(o).hexdigest()

        # Just skip silently if no match is found
        if not os.path.exists(d):
            return True

        # Patch font
        self.addLog('sPatchFont')
        return self.patchFile(self.fontname, d + '/patch', o, selected)

    def patchBSAs(self, path):
        # Check if BSA datetimes are correct
        bsa_correct = { 'Morrowind.bsa':1024669906,
                        'Tribunal.bsa':1035919326,
                        'Bloodmoon.bsa':1051789050 }

        t = map(lambda x: os.stat(path + '/' + x[0]).st_mtime == x[1], bsa_correct.iteritems())

        if not all(t):
            self.addLog('sBSARedating')
            try:
                for bsa, t in bsa_correct.iteritems():
                    os.utime(path + '/' + bsa, (t, t))
            except IOError, inst:
                self.addLog('sIOError')
                self.addLog(inst.strerror)
                return False
            
        self.addLog('sBSACorrect')

        # Check if ESM datetimes are correct
        # Disabled due to variations in localized versions
        
        #esm_correct = { 'Morrowind.esm':1026925162,
        #                'Tribunal.esm':1056650706,
        #                'Bloodmoon.esm':1057591676 }

        #t = map(lambda x: os.stat(path + '/' + x[0]).st_mtime == x[1], esm_correct.iteritems())

        #if not all(t):
        #    self.addLog('sESMRedating')
        #    try:
        #        for esm, t in esm_correct.iteritems():
        #            os.utime(path + '/' + esm, (t, t))
        #    except IOError, inst:
        #        self.addLog('sIOError')
        #        self.addLog(inst.strerror)
        #        return False
        #    
        #self.addLog('sESMCorrect')

        return True
        
    def patchFile(self, filename, patchname, o, selected):
        # Start patching
        self.addLog('sPatchApply')

        # Apply selected patches
        f = file(patchname, 'rb')
        p = array.array('B', f.read())
        f.close()
        
        offset = 0
        while offset < len(p):
            idx, base, dist = struct.unpack_from('III', p, offset)
            offset = offset + struct.calcsize('III')
            if selected.get(idx, False):
                o[base:base+dist] = p[offset:offset+dist]
            offset = offset + dist
            
        self.addLog('sPatchWrite')
        try:
            f = file(filename, 'wb')
            o.tofile(f)
            f.close()
            pass
        except IOError, inst:
            self.addLog('sIOError')
            self.addLog(inst.strerror)
            return False

        return True

    def onUninstall(self):
        mwname = 'Morrowind.exe'
        backupname = 'Morrowind.Original.exe'
        self.clearLog()
        
        # Detect Morrowind
        self.addLog('sUninstallStart')

        if not os.path.exists(mwname):
            self.addLog('sMorrowindNotFound')
            self.addLog('sUninstallFail')
            return False

        try:
            if os.path.exists(backupname):
                f = file(backupname, 'rb')
                o = f.read()
                f.close()
                f2 = file(mwname, 'wb')
                f2.write(o)
                f2.close()
            else:
                self.addLog('sBackupNotFound')
                self.addLog('sUninstallFail')
                return False
        except IOError, inst:
            self.addLog('sIOError')
            self.addLog(inst.strerror)
            self.addLog('sUninstallFail')
            return False

        dllname = 'Text.dll'
        backupname = 'Text.Original.dll'
        
        try:
            if os.path.exists(backupname):
                f = file(backupname, 'rb')
                o = f.read()
                f.close()
                f2 = file(dllname, 'wb')
                f2.write(o)
                f2.close()
        except IOError, inst:
            self.addLog('sIOError')
            self.addLog(inst.strerror)
            self.addLog('sUninstallFail')
            return False
                
        self.addLog('sUninstallSuccess')
        self.installed = {}
        return True
        
    def clearLog(self):
        self.log = []

    def addLog(self, s):
        self.log.append(s)
        self.frame.displayLog(self.log)

mcp = MCPatchApp()
